@@ -71,7 +71,7 @@ import logging
 import types
 
 from zim.fs import Dir, File
-from zim.parsing import link_type, is_url_re, \
+from zim.parsing import link_type, is_url_re, is_www_link_re, \
 	url_encode, url_decode, URL_ENCODE_READABLE, URL_ENCODE_DATA
 from zim.parser import Builder
 from zim.config import data_file, ConfigDict
@@ -107,6 +107,7 @@ UNCHECKED_BOX = 'unchecked-box'
 CHECKED_BOX = 'checked-box'
 XCHECKED_BOX = 'xchecked-box'
 MIGRATED_BOX = 'migrated-box'
+TRANSMIGRATED_BOX = "transmigrated-box"
 BULLET = '*' # FIXME make this 'bullet'
 
 FORMATTEDTEXT = 'zim-tree'
@@ -170,6 +171,25 @@ def increase_list_iter(listiter):
 		except IndexError: # wrap to start of list
 			return _letters[0]
 
+
+def convert_list_iter_letter_to_number(listiter):
+	'''Convert a "letter" numbered list to a digit numbered list
+	Usefull for export to formats that do not support letter lists.
+	Both "A." and "a." convert to "1." assumption is that this function
+	is used for start iter only, not whole list
+	'''
+	try:
+		i = int(listiter)
+		return listiter
+	except ValueError:
+		try:
+			i = _letters.index(listiter) + 1
+			i = i if i <= 26 else i % 26
+			return str(i)
+		except ValueError: # listiter is not a letter
+			return None
+
+
 def encode_xml(text):
 	'''Encode text such that it can be used in xml
 	@param text: label text as string
@@ -200,6 +220,10 @@ def canonical_name(name):
 		return name
 
 
+_aliases = {
+	'zim-wiki': 'wiki',
+}
+
 def get_format(name):
 	'''Returns the module object for a specific format.'''
 	# If this method is removes, class names in formats/*.py can be made more explicit
@@ -213,6 +237,7 @@ def get_format_module(name):
 	@param name: the format name
 	@returns: a module object
 	'''
+	name = _aliases.get(name, name)
 	return zim.plugins.get_module('zim.formats.' + canonical_name(name))
 
 
@@ -244,6 +269,12 @@ def get_dumper(name, *arg, **kwarg):
 	return klass(*arg, **kwarg)
 
 
+def heading_to_anchor(name):
+	"""Derive an anchor name from a heading"""
+	name = re.sub(r'\s', '-', name.strip().lower())
+	return re.sub(r'[^\w\-_]', '', name)
+
+
 class ParseTree(object):
 	'''Wrapper for zim parse trees.'''
 
@@ -255,7 +286,7 @@ class ParseTree(object):
 	def __init__(self, *arg, **kwarg):
 		self._etree = ElementTreeModule.ElementTree(*arg, **kwarg)
 		self._object_cache = {}
-		self.meta = OrderedDict()
+		self.meta = DefinitionOrderedDict()
 
 	@property
 	def hascontent(self):
@@ -279,6 +310,15 @@ class ParseTree(object):
 		'''
 		return self._etree.getroot().attrib.get('raw', False)
 
+	def _set_root_attrib(self, key, value):
+		self._etree.getroot().attrib[key] = value
+
+	def _get_root_attrib(self, key, default=None):
+		return self._etree.getroot().attrib.get(key, default)
+
+	def _pop_root_attrib(self, key, default=None):
+		return self._etree.getroot().attrib.pop(key, default)
+
 	def extend(self, tree):
 		# Do we need a deepcopy here ?
 		myroot = self._etree.getroot()
@@ -335,24 +375,39 @@ class ParseTree(object):
 		self.visit(tb)
 		return iter(tb.tokens)
 
-	def iter_href(self):
+	def iter_href(self, include_page_local_links=False, include_anchors=False):
 		'''Generator for links in the text
+		@param include_anchors: if C{False} remove the target location from the
+		link and only yield unique links to pages
 		@returns: yields a list of unique L{HRef} objects
 		'''
 		from zim.notebook.page import HRef # XXX
+
 		seen = set()
 		for elt in itertools.chain(
 			self._etree.iter(LINK),
 			self._etree.iter(IMAGE)
 		):
 			href = elt.attrib.get('href')
-			if href and href not in seen:
-				seen.add(href)
-				if link_type(href) == 'page':
-					try:
-						yield HRef.new_from_wiki_link(href)
-					except ValueError:
-						pass
+			if not href or link_type(href) != 'page':
+				continue
+
+			try:
+				href_obj = HRef.new_from_wiki_link(href)
+			except ValueError:
+				continue
+
+			if not include_anchors:
+				if not href_obj.names:
+					continue # internal link within same page
+				elif href_obj.anchor:
+					href_obj.anchor = None
+					href = href_obj.to_wiki_link()
+
+			if href in seen:
+				continue
+			seen.add(href)
+			yield href_obj
 
 	def iter_tag_names(self):
 		'''Generator for tags in the page content
@@ -467,7 +522,7 @@ class ParseTree(object):
 		else:
 			for element in self._etree.iter('img'):
 				filepath = element.attrib['src']
-				element.attrib['_src_file'] = notebook.resolve_file(element.attrib['src'], path)
+				element.attrib['_src_file'] = notebook.resolve_file(filepath, path)
 
 	def unresolve_images(self):
 		'''Undo effect of L{resolve_images()}, mainly intended for
@@ -483,7 +538,7 @@ class ParseTree(object):
 		'''
 		for link in self._etree.iter('link'):
 			href = link.attrib['href']
-			if href and is_url_re.match(href):
+			if href and is_url_re.match(href) or is_www_link_re.match(href):
 				link.attrib['href'] = url_encode(href, mode=mode)
 				if link.text == href:
 					link.text = link.attrib['href']
@@ -494,7 +549,7 @@ class ParseTree(object):
 		'''
 		for link in self._etree.iter('link'):
 			href = link.attrib['href']
-			if href and is_url_re.match(href):
+			if href and is_url_re.match(href) or is_www_link_re.match(href):
 				link.attrib['href'] = url_decode(href, mode=mode)
 				if link.text == href:
 					link.text = link.attrib['href']
@@ -589,11 +644,11 @@ class ParseTree(object):
 		for elt in self._etree.iter(tag):
 			yield Element.new_from_etree(elt)
 
-	def replace(self, tag, func):
+	def replace(self, tags, func):
 		'''Modify the tree by replacing all occurences of C{tag}
 		by the return value of C{func}.
 
-		@param tag: tag name
+		@param tags: tag name, or list of tag names
 		@param func: function to generate replacement values.
 		Function will be called as::
 
@@ -607,17 +662,19 @@ class ParseTree(object):
 		If the function raises L{VisitorStop} the replacement of all
 		nodes will stop.
 		'''
+		if not isinstance(tags, (tuple, list)):
+			tags = (tags,)
 		try:
-			self._replace(self._etree.getroot(), tag, func)
+			self._replace(self._etree.getroot(), tags, func)
 		except VisitorStop:
 			pass
 
-	def _replace(self, elt, tag, func):
+	def _replace(self, elt, tags, func):
 		# Two-step replace in order to do items in order
 		# of appearance.
 		replacements = []
 		for i, child in enumerate(elt):
-			if child.tag == tag:
+			if child.tag in tags:
 				try:
 					replacement = func(Element.new_from_etree(child))
 				except VisitorSkip:
@@ -625,11 +682,10 @@ class ParseTree(object):
 				else:
 					replacements.append((i, child, replacement))
 			elif len(child):
-				self._replace(child, tag, func) # recurs
+				self._replace(child, tags, func) # recurs
 			else:
 				pass
 
-
 		if replacements:
 			self._do_replace(elt, replacements)
 
@@ -637,7 +693,7 @@ class ParseTree(object):
 		offset = 0 # offset due to replacements
 		for i, child, node in replacements:
 			i += offset
-			if node is None or len(node) == 0:
+			if node is None:
 				# Remove element
 				tail = child.tail
 				elt.remove(child)
@@ -854,7 +910,8 @@ class OldParseTreeBuilder(object):
 	also be used on other "dirty" interfaces.
 
 	This builder takes care of the following issues:
-		- Inline tags ('emphasis', 'strong', 'h', etc.) can not span multiple lines
+		- ~~Inline tags ('emphasis', 'strong', 'h', etc.) can not span multiple lines~~
+		  (refactored out to `TextBuffer.get_parsetree()`)
 		- Tags can not contain only whitespace
 		- Tags can not be empty (with the exception of the 'img' tag)
 		- There should be an empty line before each 'h', 'p' or 'pre'
@@ -972,6 +1029,7 @@ class OldParseTreeBuilder(object):
 
 		#~ print('DATA:', self._data)
 		text = ''.join(self._data)
+		self._data = []
 
 		# Fix trailing newlines
 		if text:
@@ -1005,45 +1063,12 @@ class OldParseTreeBuilder(object):
 
 		if text:
 			assert not self._last is None, 'data seen before root element'
-			self._data = []
-
-			# Tags that are not allowed to have newlines
-			if not self._tail and self._last.tag in (
-			'h', 'emphasis', 'strong', 'mark', 'strike', 'code'):
-                # assume no nested tags in these types ... XXE3rd: WHY we assume no nested tags? They are now all of them nested and it still works well without change.
-				if self._seen_eol:
-					text = text.rstrip('\n')
-					self._data.append('\n' * self._seen_eol)
-					self._seen_eol = 0
-				lines = text.split('\n')
-
-				for line in lines[:-1]:
-					assert self._last.text is None, "internal error (text)"
-					assert self._last.tail is None, "internal error (tail)"
-					if line and not line.isspace():
-						self._last.text = line
-						self._last.tail = '\n'
-						attrib = self._last.attrib.copy()
-						self._last = ElementTreeModule.Element(self._last.tag, attrib)
-						self._stack[-2].append(self._last)
-						self._stack[-1] = self._last
-					else:
-						self._append_to_previous(line + '\n')
-
-				assert self._last.text is None, "internal error (text)"
-				self._last.text = lines[-1]
+			if self._tail:
+				assert self._last.tail is None, "internal error (tail)"
+				self._last.tail = text
 			else:
-				# TODO split paragraphs
-
-				if self._tail:
-					assert self._last.tail is None, "internal error (tail)"
-					self._last.tail = text
-				else:
-					assert self._last.text is None, "internal error (text)"
-					self._last.text = text
-		else:
-			self._data = []
-
+				assert self._last.text is None, "internal error (text)"
+				self._last.text = text
 
 	def close(self):
 		assert len(self._stack) == 0, 'missing end tags'
@@ -1096,7 +1121,7 @@ class ParserClass(object):
 					break
 
 				k, v = option.split('=', 1)
-				if k in ('width', 'height', 'type', 'href'):
+				if k in ('id', 'width', 'height', 'type', 'href'):
 					if len(v) > 0:
 						value = url_decode(v, mode=URL_ENCODE_DATA)
 						attrib[str(k)] = value # str to avoid unicode key
@@ -1504,7 +1529,7 @@ class Node(list):
 		if self.attrib:
 			strings.append('<%s' % self.tag)
 			for key in sorted(self.attrib):
-				strings.append(' %s="%s"' % (key, encode_xml(self.attrib[key])))
+				strings.append(' %s="%s"' % (key, encode_xml(str(self.attrib[key]))))
 			strings.append('>')
 		else:
 			strings.append("<%s>" % self.tag)
@@ -1695,7 +1720,7 @@ class TableParser():
 		return cells
 
 
-from zim.config.dicts import OrderedDict
+from zim.config.dicts import DefinitionOrderedDict
 
 _is_header_re = re.compile('^([\w\-]+):\s+(.*?)\n', re.M)
 _is_continue_re = re.compile('^([^\S\n]+)(.+?)\n', re.M)
@@ -1711,7 +1736,7 @@ def parse_header_lines(text):
 	@returns: the text minus the headers and a dict with the headers
 	'''
 	assert isinstance(text, str)
-	meta = OrderedDict()
+	meta = DefinitionOrderedDict()
 	match = _is_header_re.match(text)
 	pos = 0
 	while match:
