@@ -16,7 +16,7 @@ logger = logging.getLogger('zim.newfs')
 from . import FS_SUPPORT_NON_LOCAL_FILE_SHARES
 
 from zim.errors import Error
-from zim.parsing import url_encode
+from zim.parsing import url_encode, url_decode
 
 
 is_url_re = re.compile('^\w{2,}:/')
@@ -112,7 +112,7 @@ def _split_file_url(url):
 	else:
 		isshare = False # either 'file:/' or 'file:///'
 
-	return path.strip('/').split('/'), isshare
+	return url_decode(path).strip('/').split('/'), isshare
 
 
 def _splitnormpath(path, force_rel=False):
@@ -213,6 +213,15 @@ def _os_expanduser(path):
 			return path
 
 
+def is_abs_filepath(string):
+	try:
+		_joinabspath(_splitnormpath(string))
+	except ValueError:
+		return False
+	else:
+		return True
+
+
 class FilePath(object):
 	'''Class to represent filesystem paths and the base class for all
 	file and folder objects. Contains methods for file path manipulation.
@@ -587,6 +596,8 @@ IMAGE_EXTENSIONS = (
 	'svg', # image/svg+xml
 	'svgz', # image/svg+xml
 	'svg.gz', # image/svg+xml
+	# Custom additions
+	'webp', # image/webp
 )
 
 
@@ -654,7 +665,10 @@ class File(FSObjectBase):
 	def size(self):
 		raise NotImplementedError
 
-	def read(self):
+	def read(self, size=-1):
+		raise NotImplementedError
+
+	def readline(self, size=-1):
 		raise NotImplementedError
 
 	def readlines(self):
