@@ -10,6 +10,7 @@ logger = logging.getLogger('zim.notebook')
 
 
 from zim.parsing import link_type
+from zim.errors import Error
 
 import zim.formats
 import zim.fs
@@ -331,18 +332,18 @@ HREF_REL_RELATIVE = 2
 
 class HRef(object):
 
-	__slots__ = ('rel', 'names')
+	__slots__ = ('rel', 'names', 'anchor')
 
 	@classmethod
 	def new_from_wiki_link(klass, href):
 		'''Constructor that constructs a L{HRef} object for a link as
-		writen in zim's wiki syntax.
+		written in zim's wiki syntax.
 		@param href: a string for the link
 		@returns: a L{HRef} object
 		@raises ValueError: when the string could not be parsed
 		(see L{Path.makeValidPageName()})
 
-		@note: This mehtod HRef class assumes the logic of our wiki links
+		@note: This method HRef class assumes the logic of our wiki links
 		for other formats, a separate constructor may be needed
 		'''
 		if href.startswith(':'):
@@ -352,30 +353,37 @@ class HRef(object):
 		else:
 			rel = HREF_REL_FLOATING
 
-		names = Path.makeValidPageName(href.lstrip('+'))
-			# Can raise ValueError if link would reduce to empty string
-		return klass(rel, names)
+		anchor = None
+		if '#' in href:
+			href, anchor = href.split('#', 1)
 
-	def __init__(self, rel, names):
+		names = Path.makeValidPageName(href.lstrip('+')) if href else ""
+
+		return klass(rel, names, anchor)
+
+	def __init__(self, rel, names, anchor=None):
 		self.rel = rel
 		self.names = names
+		self.anchor = anchor
 
 	def __str__(self):
 		rel = {HREF_REL_ABSOLUTE: 'abs', HREF_REL_FLOATING: 'float', HREF_REL_RELATIVE: 'rel'}[self.rel]
-		return '<%s: %s %s>' % (self.__class__.__name__, rel, self.names)
+		return '<%s: %s %s %s>' % (self.__class__.__name__, rel, self.names, self.anchor)
 
 	def parts(self):
-		return self.names.split(':')
+		return self.names.split(':') if self.names else []
 
 	def to_wiki_link(self):
 		'''Returns href as text for wiki link'''
 		if self.rel == HREF_REL_ABSOLUTE:
-			return ":" + self.names.strip(':')
+			link = ":" + self.names.strip(':')
 		elif self.rel == HREF_REL_RELATIVE:
-			return "+" + self.names
+			link = "+" + self.names
 		else:
-			return self.names
-
+			link = self.names
+		if self.anchor:
+			link += "#" + self.anchor
+		return link
 
 
 class SourceFile(zim.fs.File):
@@ -390,6 +398,10 @@ class SourceFile(zim.fs.File):
 		raise AssertionError('Not writeable')
 
 
+class PageReadOnlyError(Error):
+	_msg = _('Can not modify page: %s') # T: error message for read-only pages
+
+
 class Page(Path, SignalEmitter):
 	'''Class to represent a single page in the notebook.
 
@@ -409,38 +421,37 @@ class Page(Path, SignalEmitter):
 	@ivar haschildren: C{True} if the page has sub-pages
 	@ivar modified: C{True} if the page was modified since the last
 	store. Will be reset by L{Notebook.store_page()}
-	@ivar readonly: C{True} when the page is read-only
-	@ivar valid: C{True} when this object is 'fresh' but C{False} e.g.
-	after flushing the notebook cache. Invalid Page objects can still
-	be used anywhere in the API where a L{Path} is needed, but not
-	for any function that actually requires a L{Page} object.
-	The way replace an invalid page object is by calling
-	C{notebook.get_page(invalid_page)}.
+	@ivar readonly: C{True} when the page is read-only or belongs to a readonly notebook
 
-	@signal: C{page-changed (changed-on-disk)}: signal emitted on page
+	@signal: C{storage-changed (changed-on-disk)}: signal emitted on page
 	change. The argument "changed-on-disk" is C{True} when an external
 	edit was detected. For internal edits it is C{False}.
+	@signal: C{modified-changed ()}: emitted when the page is edited
 	'''
 
 	__signals__ = {
-		'page-changed': (SIGNAL_NORMAL, None, (bool,))
+		'storage-changed': (SIGNAL_NORMAL, None, (bool,)),
+		'modified-changed': (SIGNAL_NORMAL, None, ()),
 	}
 
-	def __init__(self, path, haschildren, file, folder):
+	def __init__(self, path, haschildren, file, folder, format):
 		assert isinstance(path, Path)
 		self.name = path.name
 		self.haschildren = haschildren
 			# Note: this attribute is updated by the owning notebook
 			# when a child page is stored
-		self.valid = True
-		self.modified = False
+		self._modified = False
+		self._change_counter = 0
 		self._parsetree = None
-		self._ui_object = None
+		self._textbuffer = None
 		self._meta = None
 
 		self._readonly = None
 		self._last_etag = None
-		self.format = zim.formats.get_format('wiki') # TODO make configurable
+		if isinstance(format, str):
+			self.format = zim.formats.get_format(format)
+		else:
+			self.format = format
 		self.source = SourceFile(file.path) # XXX
 		self.source_file = file
 		self.attachments_folder = folder
@@ -462,17 +473,40 @@ class Page(Path, SignalEmitter):
 	@property
 	def hascontent(self):
 		'''Returns whether this page has content'''
-		if self._parsetree:
+		if self._textbuffer:
+			return self._textbuffer.hascontent
+		elif self._parsetree:
 			return self._parsetree.hascontent
-		elif self._ui_object:
-			tree = self._ui_object.get_parsetree()
-			if tree:
-				return tree.hascontent
-			else:
-				return False
 		else:
 			return self.source_file.exists()
 
+	@property
+	def modified(self):
+		return self._modified
+
+	def set_modified(self, modified):
+		if modified:
+			# HACK: by setting page.modified to a number rather than a
+			# bool we can use this number to check against race conditions
+			# in notebook.store_page_async post handler
+			self._change_counter = max(1, (self._change_counter + 1) % 1000)
+			self._modified = self._change_counter
+			assert bool(self._modified) is True, 'BUG in counter'
+		else:
+			self._modified = False
+		self.emit('modified-changed')
+
+	def on_buffer_modified_changed(self, buffer):
+		# one-way traffic, set page modified after modifying the buffer
+		# but do not set page.modified False again when buffer goes
+		# back to un-modified. Reason is that we use the buffer modified
+		# state to track if we already requested the parse tree (see
+		# get_parsetree()) while page modified is used to track need
+		# for saving and is reset after save was done
+		if buffer.get_modified():
+			if self.readonly:
+				logger.warn('Buffer edited while page read-only - potential bug')
+			self.set_modified(True)
 
 	def _store(self):
 		tree = self.get_parsetree()
@@ -504,32 +538,31 @@ class Page(Path, SignalEmitter):
 			self.source_file.remove()
 			self._last_etag = None
 			self._meta = None
+		self.emit('storage-changed', False)
 
 	def check_source_changed(self):
-		return self._check_source_etag()
+		'''Checks for changes in the source file and load it if needed
 
-	def _check_source_etag(self):
+		If the page has a C{textbuffer} and it contains unsaved changes, this
+		method will not overwrite them and you'll get an error on next attempt
+		to save. To force overwrite see L{reload_textbuffer()}
+		'''
 		if (
 			self._last_etag
-			and not self.source_file.verify_etag(self._last_etag)
+			and not (self.source_file.exists() and self.source_file.verify_etag(self._last_etag))
 		) or (
 			not self._last_etag
-			and (self._parsetree or self._ui_object)
 			and self.source_file.exists()
 		):
 			logger.info('Page changed on disk: %s', self.name)
 			self._last_etag = None
 			self._meta = None
-			self._parsetree = None
-			if self._ui_object is not None:
-				obj = self._ui_object
-				self._ui_object = None
-				parsetree = self.get_parsetree()
-				self._ui_object = obj
+			if self._textbuffer and not self._textbuffer.get_modified():
+				self.reload_textbuffer()
+			else:
 				self._parsetree = None
-				self._ui_object.set_parsetree(parsetree)
 
-			self.emit('page-changed', True)
+			self.emit('storage-changed', True)
 			return True
 		else:
 			return False
@@ -561,12 +594,14 @@ class Page(Path, SignalEmitter):
 
 		@returns: a L{zim.formats.ParseTree} object or C{None}
 		'''
-		assert self.valid, 'BUG: page object became invalid'
-
-		if self._parsetree:
+		if self._textbuffer:
+			if self._textbuffer.get_modified() or self._parsetree is None:
+				self._parsetree = self._textbuffer.get_parsetree()
+				self._textbuffer.set_modified(False)
+			#~ print self._parsetree.tostring()
+			return self._parsetree
+		elif self._parsetree:
 			return self._parsetree
-		elif self._ui_object:
-			return self._ui_object.get_parsetree()
 		else:
 			try:
 				text, self._last_etag = self.source_file.read_with_etag()
@@ -589,47 +624,87 @@ class Page(Path, SignalEmitter):
 		needs to be stored in the notebook to save this content
 		permanently. See L{Notebook.store_page()}.
 		'''
-		assert self.valid, 'BUG: page object became invalid'
-
 		if self.readonly:
 			raise PageReadOnlyError(self)
+		self._set_parsetree(tree)
 
-		if self._ui_object:
-			self._ui_object.set_parsetree(tree)
-		else:
-			self._parsetree = tree
+	def _set_parsetree(self, tree):
+		self._parsetree = tree
+		if self._textbuffer:
+			assert not self._textbuffer.get_modified(), 'BUG: changing parsetree while buffer was changed as well'
+			try:
+				if tree is None:
+					self._textbuffer.clear()
+				else:
+					self._textbuffer.set_parsetree(tree)
+			except:
+				# Prevent auto-save to kick in at any cost
+				self._textbuffer.set_modified(False)
+				raise
+			else:
+				self._textbuffer.set_modified(False)
 
-		self.modified = True
+		self.set_modified(True)
 
 	def append_parsetree(self, tree):
 		'''Append content
 
 		@param tree: a L{zim.formats.ParseTree} object with content
 		'''
-		ourtree = self.get_parsetree()
-		if ourtree:
-			self.set_parsetree(ourtree + tree)
+		if self._textbuffer:
+			self._textbuffer.append_parsetree(tree)
 		else:
-			self.set_parsetree(tree)
+			ourtree = self.get_parsetree()
+			if ourtree:
+				self.set_parsetree(ourtree + tree)
+			else:
+				self.set_parsetree(tree)
 
-	def set_ui_object(self, object):
-		'''Lock the page to an interface widget
+	def get_textbuffer(self, constructor=None):
+		'''Get a C{Gtk.TextBuffer} for the page
 
-		Setting a "ui object" locks the page and turns it into a proxy
-		for that widget - typically a L{zim.gui.pageview.PageView}.
-		The "ui object" should in turn have a C{get_parsetree()} and a
-		C{set_parsetree()} method which will be called by the page object.
+		Will either return an existing buffer or construct a new one and return
+		it. A C{Gtk.TextBuffer} can be shared between multiple C{Gtk.TextView}s.
+		The page object owns the textbuffer to allow multiple views on the same
+		page.
 
-		@param object: a widget or similar object or C{None} to unlock
+		Once a buffer is set, also methods like L{get_parsetree()} and
+		L{get_parsetree()} will interact with this buffer.
+
+		@param constructor: if not buffer was set previously, this function
+		is called to construct the buffer.
+
+		@returns: a C{TextBuffer} object or C{None} if no buffer is set and
+		no constructor is provided.
 		'''
-		if object is None:
-			if self._ui_object:
-				self._parsetree = self._ui_object.get_parsetree()
-				self._ui_object = None
-		else:
-			assert self._ui_object is None, 'BUG: page already being edited by another widget'
-			self._parsetree = None
-			self._ui_object = object
+		if self._textbuffer is None:
+			if constructor is None:
+				return None
+
+			tree = self.get_parsetree()
+			self._textbuffer = constructor(parsetree=tree)
+			self._textbuffer.connect('modified-changed', self.on_buffer_modified_changed)
+
+		return self._textbuffer
+
+	def reload_textbuffer(self):
+		'''Reload page content from source file and update the textbuffer if set
+
+			NOTE: this method overwrites any changes in the C{textbuffer} or
+			C{parsetree} that have not been saved to file !
+		'''
+		buffer = self._textbuffer
+		self._textbuffer = None
+		self._parsetree = None
+		if buffer is not None:
+			tree = self.get_parsetree()
+			self._textbuffer = buffer
+			buffer.set_modified(False)
+			self._set_parsetree(tree)
+				# load new tree in buffer, undo-able in 1 step
+				# private method circumvents readonly check !
+			self.set_modified(False)
+		# else do nothing - source will be read with next call to `get_parsetree()`
 
 	def dump(self, format, linker=None):
 		'''Get content in a specific format
@@ -723,6 +798,17 @@ class Page(Path, SignalEmitter):
 					seen.add(name)
 					yield name.lstrip('@'), elt.attrib
 
+	def get_anchors(self):
+		'''Generator returning all the (explicit) anchors in the page content'''
+		tree = self.get_parsetree()
+		if tree:
+			seen = set()
+			for elt in tree.findall(zim.formats.ANCHOR):
+				name = elt.gettext()
+				if not name in seen:
+					seen.add(name)
+					yield name, elt.attrib
+
 	def get_title(self):
 		tree = self.get_parsetree()
 		if tree:
